/***************************************************************************
 *   Copyright (C) 2005 by Jaymin Kessler                                  *
 *   root@localhost.localdomain                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "DJR_FrameMemMngr.h"

// statics 
static int m_iAlignment;
static MemStack m_Stack[ STACKTYPE_NUMTYPES ];
static size_t m_iBytesPerStack;
static unsigned long m_iApproxTotalMemUsed;

/***************************************************************************/
void FrameMemMngrInit( void )
{
  int iStack;
	
  m_iAlignment = -1;
  m_iBytesPerStack = 0;
  m_iApproxTotalMemUsed = 0;
  // init all pointers
  for(iStack=0; iStack<STACKTYPE_NUMTYPES; iStack++ )
  { 
    FrameMemMngrResetAllPointers( iStack );
  }
}

/***************************************************************************/		
void FrameMemMngrKill( void )
{
  // look for memory leaks in case KillMemMngr() wasnt called before exit
  int iStack;
	
  // yes, I am aware that I dont need to reset most of these class variables
  m_iAlignment = -1;
  m_iBytesPerStack = 0;
  m_iApproxTotalMemUsed = 0;

  for(iStack=0; iStack<STACKTYPE_NUMTYPES; iStack++ )
  {
    if( m_Stack[iStack].pPermanentMemPtr != NULL )
    {
      printf( "\nuh oh.  Someone forgot to call KillMemMngr() before exiting.  I guess I can clean up your mess..." );
      free( m_Stack[ iStack].pAddrToFree );  
    }
    FrameMemMngrResetAllPointers( iStack );
  }
}

/***************************************************************************/
void FrameMemMngrResetAllPointers( int iStack )
{
  //m_pPermanentMemPtr[ iStack ] = NULL; //oops, dont set this to NULL if memory is allocated
  m_Stack[ iStack].pStackPtr[ STACK_PTR_LOWER ] = NULL;         
  m_Stack[ iStack].pStackPtr[ STACK_PTR_UPPER ] = NULL;
}

/***************************************************************************/
Bool FrameMemMngrCreate( int iNumBytes, int iAlignment )
{
  int iStack;
	
  m_iAlignment = iAlignment;
	
  // TODO: make sure iAlignment is a power of 2
	
  // min amount of memory we can allocate
  iNumBytes = iNumBytes < m_iAlignment? m_iAlignment : iNumBytes;

  printf( "\n\ninitializing memory manager:" );
  printf( "\n   num stacks: %d", STACKTYPE_NUMTYPES );
  printf( "\n   address alignment: %d", m_iAlignment );
  printf( "\n   bytes per stack before boundary correction: %d (multiple of %d=%s)", iNumBytes, m_iAlignment, iNumBytes%m_iAlignment==0?"yes":"no" );
		
  // example: we want to allocate 75 bytes aligned on a 8 byte boundary
  // 75 + (8 - (75 % 8)) = 80 bytes
  iNumBytes = ALIGN_MEMORY( iNumBytes, m_iAlignment, int );
  m_iBytesPerStack = iNumBytes;
  printf( "\n   bytes per stack after boundary correction: %d (multiple of %d=%s)", iNumBytes, m_iAlignment, iNumBytes%m_iAlignment==0?"yes":"no" );

  m_iApproxTotalMemUsed = m_iBytesPerStack * STACKTYPE_NUMTYPES;
  printf( "\n   approximate memory used by stacks: %ld bytes", m_iApproxTotalMemUsed );

  // allocate the actual stack memory
  for( iStack=0; iStack<STACKTYPE_NUMTYPES; iStack++ )
  {		
    //if( !posix_memalign( &m_Stack[ iStack].pPermanentMemPtr, m_iAlignment, iNumBytes ) )
    if( m_Stack[ iStack ].pAddrToFree = malloc( iNumBytes + m_iAlignment - 1 ) )
    {
      printf( "\n\nstack %d:", iStack );
      printf( "\n   address returned by malloc: %X (multiple of %d=%s)", m_Stack[ iStack ].pAddrToFree, m_iAlignment, 
	                                                                 (int)m_Stack[ iStack ].pAddrToFree%m_iAlignment==0?"yes":"no" );
      // take the malloc'd pointer and move it to the proper alignment
      m_Stack[ iStack ].pPermanentMemPtr = ALIGN_MEMORY( m_Stack[ iStack].pAddrToFree, m_iAlignment, void* );
      printf( "\n   corrected address: %X (multiple of %d=%s)", m_Stack[ iStack ].pPermanentMemPtr, m_iAlignment,
	                                                        (int)m_Stack[ iStack ].pPermanentMemPtr%m_iAlignment==0?"yes":"no" );
      // set up the stack
      m_Stack[ iStack ].pStackPtr[ STACK_PTR_LOWER ] = m_Stack[ iStack ].pPermanentMemPtr;
      printf( "\n   lower stack pointer: %X (multiple of %d=%s)", m_Stack[ iStack ].pStackPtr[ STACK_PTR_LOWER ], m_iAlignment,
	                                                          (int)m_Stack[ iStack ].pStackPtr[ STACK_PTR_LOWER ]%m_iAlignment==0?"yes":"no" );
      m_Stack[ iStack ].pStackPtr[ STACK_PTR_UPPER ] = (char*)m_Stack[ iStack ].pPermanentMemPtr + iNumBytes;
      printf( "\n   upper stack pointer: %X (multiple of %d=%s)", m_Stack[ iStack ].pStackPtr[ STACK_PTR_UPPER ], m_iAlignment,
	                                                          (int)m_Stack[ iStack ].pStackPtr[ STACK_PTR_UPPER ]%m_iAlignment==0?"yes":"no" );
    }
    else
    {
      // oooooh man something bad is going on.  Lets get outta here
      // TODO: get outta here
      m_Stack[ iStack ].pPermanentMemPtr = NULL;
      m_Stack[ iStack ].pAddrToFree = NULL;
      printf( "\n\nstack %d failed to allocate %d bytes", iStack, iNumBytes + m_iAlignment - 1 );
      return FALSE;
    }
  }
  fflush( NULL );
  return TRUE;
}

/*************************************************************************/
void FrameMemMngrDestroy( void )
{
  int iStack;

  for( iStack=0; iStack<STACKTYPE_NUMTYPES; iStack++ )
  {
    if( m_Stack[ iStack ].pPermanentMemPtr )
    {
      free( m_Stack[ iStack ].pAddrToFree );
      m_Stack[ iStack ].pPermanentMemPtr = NULL;
    }
    FrameMemMngrResetAllPointers( iStack );
  }
}

/***************************************************************************/
void *FrameMemMngrAllocate( StackType type, Bool bUpper, int iNumBytes )
{
  void *pRetAddr = NULL;
	
  if( bUpper )
  {
    // check to see if there is enough memory for this allocation
    if( (char*)m_Stack[ type ].pStackPtr[ STACK_PTR_UPPER ] - iNumBytes >= m_Stack[ type ].pStackPtr[ STACK_PTR_LOWER ] )
    {
      m_Stack[ type ].pStackPtr[ STACK_PTR_UPPER ] = (char*)m_Stack[ type ].pStackPtr[ STACK_PTR_UPPER ] - iNumBytes;
      pRetAddr = m_Stack[ type ].pStackPtr[ STACK_PTR_UPPER ];
    }
  }
  else
  {
    // check to see if there is enough memory for this allocation
    if( (char*)m_Stack[ type ].pStackPtr[ STACK_PTR_LOWER ] + iNumBytes <= m_Stack[ type ].pStackPtr[ STACK_PTR_UPPER ] )
    {
      m_Stack[ type ].pStackPtr[ STACK_PTR_LOWER ] = (char*)m_Stack[ type ].pStackPtr[ STACK_PTR_LOWER ] + iNumBytes;
      pRetAddr = m_Stack[ type ].pStackPtr[ STACK_PTR_LOWER ];
    }
  }

  if( !pRetAddr )
    printf( "\n\nFrameMemMngr: memory stack %d doesnt have enough free memory to allocate %d bytes" ,type, iNumBytes );
	
  return pRetAddr;
}

/*************************************************************************/
void *FrameMemMngrObtainFrame( StackType type, Bool bUpper )
{
  return( m_Stack[ type ].pStackPtr[ bUpper ] );
}

/*************************************************************************/
Bool FrameMemMngrDeallocateFrame( StackType type, Bool bUpper, void *pFrame )
{
  // check to see if the frame makes sense 
  if( ( bUpper && pFrame < m_Stack[ type ].pStackPtr[ STACK_PTR_LOWER ] ) || 
      ( !bUpper && pFrame > m_Stack[ type ].pStackPtr[ STACK_PTR_UPPER ] ) )
        return FALSE;
		
  m_Stack[ type ].pStackPtr[ bUpper ] = pFrame;
  return TRUE;
}

/*************************************************************************/
void FrameMemMngrDeallocateAll( StackType type, Bool bUpper )
{
  if( bUpper )
    m_Stack[ type ].pStackPtr[ STACK_PTR_UPPER ] = 
      (char*)m_Stack[ type ].pPermanentMemPtr + m_iBytesPerStack;
  else
    m_Stack[ type ].pStackPtr[ STACK_PTR_LOWER ] = 
      m_Stack[ type ].pPermanentMemPtr;
}

/*************************************************************************/


