
#include "DJR_SimpleTexMngr.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sps2lib.h>
#include <sps2tags.h>
#include <math.h>
#include "sps2util_Godai.h"
#include <sps2regstructs.h>
#include "gsregs.h"

// statics
int m_iSPS2Device;
int m_iFirstGSPage;
sps2Memory_t *m_pDMAPage;  
TextureSlot m_TextureSlots[ MAX_TEXTURE_SLOTS ];

// prototypes
int SetTexRegsForTexUpload (void *pvBase, unsigned int w, 
			    unsigned int h, 
                            sps2uint32 tex_mem_addr );
int PrepareTextureTransfer( void *pvBase, 
			    unsigned char *tex, 
			    unsigned int tex_index,
			    unsigned int qw_to_xfer );

/**************************************************/
void SimpleTexMngrInit( int iSPS2Device )
{
  m_iSPS2Device = iSPS2Device;
  m_iFirstGSPage = sps2UScreenGetFirstFreeGSPage( );
  m_pDMAPage = sps2Allocate( 4096, 
	        	     SPS2_MAP_BLOCK_4K | SPS2_MAP_UNCACHED, 
			     m_iSPS2Device );
  // setup texture slots
  memset( m_TextureSlots, 0, sizeof( TextureSlot[ 4 ] ) );
}

/**************************************************/
void SimpleTexMngrKill( void )
{
  sps2Free( m_pDMAPage );
}

/**************************************************/
sps2uint32 GetSlotGSAddress( iSlotNum )
{
  m_TextureSlots[ iSlotNum ].uGSMemAddr;
}

/**************************************************/
Bool UploadTexToGS( char *sFileName, int iRequestedSlot, 
		    TexturePriority ePriority, char *sDesc )
{
  int iQWC;
  Dn_CHCR_t chcrValue;
  int fh = 0;
  unsigned int byte_index = 0;
  unsigned int qw_to_xfer = 0;
  unsigned char *tex_data = NULL;
	
  chcrValue.i32=0;                  // Blank everything out
  chcrValue.s.MOD=CHCR_MOD_NORMAL;  // We'll do a "normal" dma transfer
  chcrValue.s.STR=1;                // STR=1 = start transfer
  chcrValue.s.DIR = 1;
	
  // load the texture from a file
  tex_data = ( unsigned char * )calloc( TEX_SIZE, sizeof( char ) );
  if( ( fh = open( sFileName, O_RDONLY ) ) == -1 )
  {
    printf( "\n\nfailed to open texture %s\n\n", sFileName );
    fflush( NULL );
    return FALSE;
  }
  read( fh, tex_data, TEX_SIZE );
	
  // set up the texture slot
  if( iRequestedSlot < 0 || iRequestedSlot >= MAX_TEXTURE_SLOTS )
  {
    // TODO: add code to look for the best open slot, 
    // or look for a loaded texture to dump
    iRequestedSlot = 0;
  }
	
  // a debugging string so I know whats in memory
  strncpy( m_TextureSlots[ iRequestedSlot ].sName, sDesc, 32 );

  // gs page * 2048 gives us the word address where texture memory starts
  m_TextureSlots[ iRequestedSlot ].uGSMemAddr = 
    ( m_iFirstGSPage * 2048 ) + ( iRequestedSlot * TEX_NUM_PIX );

  // this slot now has a texture
  m_TextureSlots[ iRequestedSlot ].bUsed = TRUE;

  // this code assumes everywhere that a textel is 32 bits everywhere
  m_TextureSlots[ iRequestedSlot ].uNumTextels = TEX_NUM_PIX;

  // how important is it to keep this texture loaded?
  m_TextureSlots[ iRequestedSlot ].ePriority = ePriority;

  // how used is this texture?
  m_TextureSlots[ iRequestedSlot ].uRefCount = 0;
	
  // set bitbltbuf, trxpos, trxreg, and trxdir
  // to set up GS texture memory
  // a page is 8K bytes or 2048 words 
  iQWC = SetTexRegsForTexUpload( m_pDMAPage->pvStart, TEX_W, TEX_H, 
			      m_TextureSlots[ iRequestedSlot ].uGSMemAddr);
	
  // dma xfer
  *EE_D2_QWC = iQWC;
  *EE_D2_MADR = sps2GetPhysicalAddress( m_pDMAPage->pvStart, m_pDMAPage );
  *EE_D2_CHCR = chcrValue.i32;
  sps2WaitForDMA( 2, m_iSPS2Device );
	
  // load the texture data - the ENTIRELY TOO SLOW way for now
  if (!(qw_to_xfer = (TEX_SIZE / 16) % MAX_TX_QW))
    qw_to_xfer = MAX_TX_QW;
	
  byte_index = 0;
  while (byte_index < TEX_SIZE)
  {
    iQWC = PrepareTextureTransfer( m_pDMAPage->pvStart, 
                                   tex_data, 
                                   byte_index, 
                                   qw_to_xfer ); 
    
    *EE_D2_QWC = iQWC;
    *EE_D2_MADR = sps2GetPhysicalAddress( m_pDMAPage->pvStart, 
                                          m_pDMAPage );
    *EE_D2_CHCR = chcrValue.i32;
    sps2WaitForDMA( 2, m_iSPS2Device );
    byte_index += qw_to_xfer * 16;
    qw_to_xfer = MAX_TX_QW; 
  }
	
  free( tex_data );
  printf( "\n\nprobably successfully uploaded %s to the GS", sFileName);
  fflush( NULL );
  return TRUE;
}

/*------------------------------------------------*/
int SetTexRegsForTexUpload(void *pvBase, unsigned int w, 
                           unsigned int h, 
                           sps2uint32 tex_mem_addr ) 
{
  sps2GIFTag_t *pGIFTag=(sps2GIFTag_t *) pvBase;
  sps2uint64 *regs = (sps2uint64 *)((char*)pvBase + sizeof(sps2GIFTag_t));
	
  // Prepare the GIF tag  
  pGIFTag->i128=0;       // Blank out the union
  pGIFTag->s.NLOOP=4;    // loop 4 times
  pGIFTag->s.EOP=1;      // end of packet
  pGIFTag->s.PRE=0;      // ignore the PRIM field
  pGIFTag->s.PRIM=0;   
  pGIFTag->s.FLG=GIF_FLG_PACKED; 
  pGIFTag->s.NREG=0;     // total registers = NREG * NLOOP
  
  // Add A+D register
  SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_A_D);
	
  /* bitbltbuf - parameters for xfer between 2 buffers
       SBP - source buffer base pointer = 0 (not used here)
       SBW - source buffer width = 0 (not used here)
       SPSM - source pixel format = 0 (32bit)
       DBP - destination buf base ptr (addr / 64) = tex_mem_addr
       DBW - destination buf width (pixels / 64) = 256 / 64
       DPSM - dest pixel format = 0
  */
  *(regs++) = PACK_BITBLTBUF(0, 0, 0, tex_mem_addr/64, TEX_W/64 , 0);
  *(regs++) = BITBLTBUF_R;
	
	
  /* trxpos - specify transmission area
     SSAX - source buffer upper left X = 0 (not used here)
     SSAY - source buffer upper left Y = 0 (not used here)
     DSAX - dest buffer upper left X = 0
     DSAY - dest buffer upper left Y = 0
     DIR - pixel xfer order = 0 (upper left -> lower right)
  */
  *(regs++) = PACK_TRXPOS(0,0,0,0,0);
  *(regs++) = TRXPOS_R;
	
	
  /* trxreg - width and height of transmission areas
     note: depends on pixel mode set in bitbltbuf
	   RRW - width of transmission area = w (num of pixels)
	   RRH - height of transmission area = h (num of pixels)
  */
  *(regs++) = PACK_TRXREG(w,h);
  *(regs++) = TRXREG_R;
	
	
  /* trxdir - transmission direction
		XDIR = 00 (begin transfering from the EE to the GS)
		*/
  *(regs++) = TRXDIR_EE_TO_GS;
  *(regs++) = TRXDIR_R;
	
  return (((int) regs -(int) pvBase) >> 4);
}

/*------------------------------------------------*/
int PrepareTextureTransfer( void *pvBase, 
                            unsigned char *tex, 
			    unsigned int tex_index,
			    unsigned int qw_to_xfer )
{
  sps2GIFTag_t *gt = (sps2GIFTag_t *)pvBase;
  gt->i128 = 0;
  gt->s.NLOOP = qw_to_xfer;
  gt->s.EOP = 1;
  gt->s.PRE = 0;
  gt->s.PRIM = 0;
  gt->s.FLG = GIF_FLG_IMAGE;
  gt->s.NREG = 0;
	
  memcpy((void*)((char*)pvBase+16),&tex[tex_index],qw_to_xfer*16);
  
  return(qw_to_xfer+1);
}

/*------------------------------------------------*/
int DrawTexturedSprite( int ulx, int uly, int llx, int lly,
                        int lrx, int lry, int urx, int ury,
                        unsigned int tw, unsigned int th, 
                        int iTextureSlot )
{
  sps2uint32 tex_mem_addr = m_TextureSlots[ iTextureSlot ].uGSMemAddr;	
  void *pvBase = m_pDMAPage->pvStart;
	
  sps2GIFTag_t *gt = (sps2GIFTag_t *)pvBase;
  sps2uint64 *regs = (sps2uint64 *) ((char*)pvBase+sizeof(sps2GIFTag_t));
  sps2GIFPackedRegister_t *pr;
	
  gt->i128 = 0;
  gt->s.NLOOP = 1;
  gt->s.EOP = 1;
  gt->s.PRE = 0;
  gt->s.PRIM = 0;
  gt->s.FLG = GIF_FLG_PACKED;
  gt->s.NREG = 0;
	
  // add an A_D register
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // TEXFLUSH
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // TEX0_1
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // PRIM
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // ALPHA_1
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // PRMODECONT

  // access the texflush register with anything
  *(regs++) = (sps2uint64)0xAA55;
  *(regs++) = TEXFLUSH_R;

	
  /* tex0_1 - texture information settings
		TBP0 - texture base pointer (word addr / 64)
		TBW - texture buffer width (textels / 64)
		PSM - pixel storage format (0 = RGBA 32)
		TW - texture width (2^tw = 256)
		TH - texture height (2^tw = 256)
		TCC - texture color component (1 = RGBA)
		TFX - texture function (1 = decal)
		CBP - CLUT buffer base ptr (word addr / 64)
		CPSM - CLUT pixel storage format = 0 = 32bit
		CSM - CLUT storage mode = CSM1
		CSA - CLUT entry offset (offset / 16)
		CLD - CLUT buffer load control = 0 (temp cont not changed)
		*/     
  int iTexWidthLog2;
  int iTexHeightLog2;
  float fExponent;
  for( fExponent = 0.f; fExponent < 16.f; fExponent += 1.f )
  {
    if( ( int )powf( 2.f, fExponent ) == TEX_W )
      iTexWidthLog2 = ( int )fExponent;
    if( ( int )powf( 2.f, fExponent ) == TEX_H )
      iTexHeightLog2 = ( int )fExponent;
  } 
  *(regs++) = PACK_TEX0(tex_mem_addr/64,TEX_W/64,0,
                        iTexWidthLog2, iTexHeightLog2,
                        1,1,0,0,0,0,0);
  *(regs++) = TEX0_1_R;
	
  /* prim - drawing primitive settings
		PRIM = drawing primitive (4 = tri strip)
		IIP - shading method (0 = flat shading)
		TME - texture mapping (1 = on)
		FGE - fogging (0 = off)
		ABE - alpha blending (1 = on)
		AA1 - 1 pass antialiasing (1 = on)
		FST - specify tex coords (0 = provide ST register later)
		CTXT - texture context (0 = use context 1)     
		FIX - fragment value control (0 = unfixed)
		*/
  *(regs++) = PACK_PRIM(4,0,1,0,1,1,0,0,0);
  *(regs++) = PRIM_R;

  // (source_rgb - 0) * source_alpha + framebuffer_rgb
  *(regs++) = PACK_ALPHA_1(0,1,0,1,105);
  *(regs++) = ALPHA_1_R;

  // PRMODECONT
  *(regs++) = ( sps2uint64 )1;
  *(regs++) = GIF_A_D_REG_PRMODECONT;
	
  // a second giftag
  gt = ( sps2GIFTag_t * )regs;
  regs += 2; 
  gt->i128 = 0;
  gt->s.NLOOP = 5;
  gt->s.EOP = 1;
  gt->s.PRE = 0;
  gt->s.PRIM = 0;
  gt->s.FLG = GIF_FLG_PACKED;
  gt->s.NREG = 0;

  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_ST);      // ST
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);     // XYZ2

  // ST - upper left
  pr = (sps2GIFPackedRegister_t *)regs;
  pr->ST.i128 = 0;
  pr->ST.s.Q = 1;
  regs += 2;
	
  // XYZ
  *(regs++) = PACK_XYZ2((2048<<4)+ulx, (2048<<4)+uly, 1024);
  *(regs++) = XYZ2_R;

  // ST - lower left
  pr = (sps2GIFPackedRegister_t *)regs;
  pr->ST.i128 = 0;
  pr->ST.s.T = 1;
  pr->ST.s.Q = 1;
  regs += 2;
	
  // XYZ
  *(regs++) = PACK_XYZ2((2048<<4)+llx, (2048<<4)+lly, 1024);
  *(regs++) = XYZ2_R;

  // ST - lower right
  pr = (sps2GIFPackedRegister_t *)regs;
  pr->ST.s.S = 1;
  pr->ST.s.T = 1;
  pr->ST.s.Q = 1;
  regs += 2;

  // XYZ2
  *(regs++) = PACK_XYZ2((2048<<4)+lrx, (2048<<4)+lry, 1024);
  *(regs++) = XYZ2_R;

  // ST - upper right
  pr = (sps2GIFPackedRegister_t *)regs;
  pr->ST.i128 = 0;
  pr->ST.s.S = 1;
  pr->ST.s.Q = 1;
  regs += 2;
	
  // XYZ2
  *(regs++) = PACK_XYZ2((2048<<4)+urx, (2048<<4)+ury, 1024);
  *(regs++) = XYZ2_R;

  // ST - upper left
  pr = (sps2GIFPackedRegister_t *)regs;
  pr->ST.i128 = 0;
  pr->ST.s.Q = 1;
  regs += 2;
	
  // XYZ
  *(regs++) = PACK_XYZ2((2048<<4)+ulx, (2048<<4)+uly, 1024);
  *(regs++) = XYZ2_R;
 
  int iQWC = (((int)regs - (int)pvBase) >> 4);
  *EE_D2_QWC = iQWC;

  Dn_CHCR_t chcrValue;
  chcrValue.i32 = 0;
  chcrValue.s.STR = 1;
  chcrValue.s.DIR = 1;
  chcrValue.s.MOD = CHCR_MOD_NORMAL;
	
  *EE_D2_MADR = sps2GetPhysicalAddress( m_pDMAPage->pvStart, m_pDMAPage );
  *EE_D2_CHCR = chcrValue.i32;

  sps2WaitForDMA( 2, m_iSPS2Device );
}
